"""
Helper Module describing functions to parse data with a html-style input from Liquipedia

Dependencies
------------
- BeautifulSoup
- pandas
- re
- mwparserfromhell: MediaWiki wikitext parsing
- parse_liquipedia: Other internal parsing utilities

Raises
------
parse_liquipedia.SectionNotFoundException
    Raised when a section like achievements or gear is not found or supported using a given method.


"""
import re
import warnings
from typing import Dict, List, Union, Optional, Tuple
from collections import defaultdict
import pandas as pd
from bs4 import BeautifulSoup
from bs4.element import Tag
from ggpyparser.parse_liquipedia import parse_liquipedia_wc


def parse_bracket_recursive_html(
    bracket: BeautifulSoup,
    subround_names: List[Tag],
    bracket_output: Optional[defaultdict] = None
) -> defaultdict:
    """
    Parses a bracket recursively
    Parameters
    ----------
        bracket: BeautifulSoup
            A beautifulsoup object describing the bracket to be parsed
        subround_names: List[Tag]
            A list of strings describing the subround names
        bracket_output: defaultdict(list)
            A dictionary mapping subround names to lists of bs4 objects describing each subround
    Returns
    -------
        bracket_output: defaultdict(list)
            A dictionary mapping subround names to lists of bs4 objects describing each subround

    """
    if bracket_output is None:
        bracket_output = defaultdict(list)

    closest_header = bracket.find_previous("div", class_="brkts-round-header")
    assert closest_header is not None
    closest_header  = [subround_name.contents[0] for subround_name in closest_header
                       if isinstance(subround_name, Tag)]
    closest_header  = [name for name in closest_header
                       if str(name).lower().strip() != "qualified"]
    #recheck headers
    if set(closest_header).isdisjoint(set(subround_names)):
        #no common between original header list and now new closest header list,
        # new header list takes over
        subround_names = closest_header

    #recursive method to parse through bracket
    names_copied = subround_names.copy()
    subbodies = []
    for child in bracket.find_all("div", recursive=False):
        #find all brkts-round-body in the next-next level
        subbodies = subbodies + child.find_all("div", class_="brkts-round-body", recursive=False)
    matches = bracket.find_all("div", class_="brkts-round-center", recursive=False)
    last_round = names_copied.pop()
    bracket_output[last_round] = bracket_output[last_round] + matches
    for div in subbodies:
        parse_bracket_recursive_html(div, names_copied, bracket_output)
    return bracket_output


def parse_side_scores_html(table: BeautifulSoup) -> List[Tuple[str, str, Optional[int]]]:
    """
    Parses the scores for a side for a map - counterstrike only so far
    Parameters
    ----------
        table: BeautifulSoup
            A beautifulsoup object describing the html of the data containing the scores
    Returns
    -------
        scores_with_halves: List[Tuple[Union[str, int]]]
            A dictionary mapping subround names to lists of bs4 objects describing each subround

    """
    scores_with_halves = []
    half_counter = 1
    ot_counter = 1

    for td in table.select("td"):
        classes = td.get("class", [])
        val = int(td.get_text(strip=True)) if td.get_text() else None

        if any("brkts-cs-score-color-" in c for c in classes):
            side = "T" if any("brkts-cs-score-color-ct" in c for c in classes) else "CT"
            label = f"Half_{half_counter}"
            if half_counter > 2:
                label = f"OT_{ot_counter}"
                ot_counter += 1
            half_counter += 1
            scores_with_halves.append((label, side, val))


    return scores_with_halves

def parse_match_html(match:BeautifulSoup) -> pd.DataFrame:
    """
    Parses a single match(multiple maps) given html
    Parameters
    ----------
        match: BeautifulSoup
            A beautifulsoup object describing the html of the data containing map information
    Returns
    -------
        pd.DataFrame
            A dataframe describing information about the match

    """
    countdown = match.find("span", class_ = "match-info-countdown")
    time = countdown.get_text() if countdown else None
    lhs_tag = match.select_one(".match-info-header-opponent-left")
    lhs_url, lhs_title = None, None
    if lhs_tag:
        name_link = lhs_tag.select_one(".name a")
        if name_link:
            lhs_url = name_link['href']
            lhs_title = name_link['title']
    rhs_url, rhs_title = None, None
    rhs_tag = match.select_one(".match-info-header-opponent:not(.match-info-header-opponent-left)")
    if rhs_tag:
        name_link = rhs_tag.select_one(".name a")
        if name_link:
            rhs_url = name_link['href']
            rhs_title = name_link['title']
    games = match.find_all("div", class_ = "brkts-popup-body-element brkts-popup-body-game")
    if len(games) == 0:
        #no games, return empty dataframe
        warnings.warn("No games found in the provided match HTML - " \
        "indicates potential forfeited match.")
        cols = ["t1","t1_url","t2","t2_url","map","t1_scores","t2_scores","time"]
        return pd.DataFrame([{}], columns=cols)
    parsed_games = []
    for game in games:
        map_tag = game.select_one("a")
        map_name = map_tag.get_text(strip=True)

        left_table = game.select_one("div[style*='direction:ltr'] table")
        left_scores = parse_side_scores_html(left_table)
        left_scores = {f"{half}_{side}": score for half, side, score in left_scores}

        right_table = game.select_one("div[style*='direction:rtl'] table")
        right_scores = parse_side_scores_html(right_table)
        right_scores = {f"{half}_{side}": score for half, side, score in right_scores}
        map_dict = {"t1": lhs_title, "t1_url": lhs_url, "t2": rhs_title, "t2_url": rhs_url,
                     "map": map_name,"t1_scores": left_scores, "t2_scores": right_scores,
                       "time": time}
        parsed_games.append(map_dict)

    return pd.DataFrame(parsed_games)

def get_all_under_header(str_bs4 : BeautifulSoup, header: str):
    """
    Gets all objects under a header but before the next header of equivalent depth
    Parameters
    ----------
        str_bs4: BeautifulSoup
            A beautifulsoup object describing the html of an entire page
        header: str
            The header of interest
    Returns
    -------
        scores_with_halves: List[Tuple[Union[str, int]]]
            A dictionary mapping subround names to lists of bs4 objects describing each subround

    """
    #parses all under a specified header until a header of the same level is reached
    header_bs4 = str_bs4.find("span", class_="mw-headline", id= header)
    data = []
    if header_bs4:
        header_tag = header_bs4.find_parent(lambda tag: tag.name is not None and
                                            tag.name.startswith("h") and tag.name[1:].isdigit()
                                            and 1 <= int(tag.name[1:]) <= 6)
        header_level = int(header_tag.name[1])  if header_tag else -1
        siblings = header_tag.find_next_siblings() if header_tag else []
        for sib in siblings:
            if isinstance(sib, Tag):
                # Check if sibling itself is a header
                if sib.name in [f"h{i}" for i in range(1, 7)]:
                    sib_level = int(sib.name[1])
                    if sib_level <= header_level:
                        break
                data.append(sib)
    return data

def parse_wikitable_hdhd(table : BeautifulSoup, larger_header : bool = True,
                         rm_1 : bool= False) -> Tuple[Dict[str, str], str]:
    """
    Parses a header-data-header-data(hdhd) style wikitable
    Parameters
    ----------
        table: BeautifulSoup
            A beautifulsoup object describing the html of the hdhd table
        larger_header: boolean
            Whether a larger, over-arching header is part of the hdhd table
        rm_1: boolean
            Whether to remove the first row
    This function is pretty bloated, TODO rewrite
    Returns
    -------
        Dict[str, str]
            A mapping between headers and their corresponding values
        str
            The overarching header of the table, if larger_header is False, this
            is the empty string

    """
    # Extract headers from the first <tr> with <th> tags
    rows = table.find_all("tr")
    total_title = ""
    if larger_header:
        total_title = rows[0].find('th')
        for tag in total_title.find_all(["sup", "small", "div"]):
            tag.decompose()
        total_title = total_title.get_text(strip=True)
        rows = rows[1:]
    prev_row_length = -1
    header_row = []
    table_data = []

    for row in rows:
        cells = [td.get_text(strip=True) for td in row.find_all(["td", "th"]) if
                 len(td.get_text(strip = True)) != 0]
        if len(cells) == prev_row_length:
            #same header
            if not (len(cells) == 1 and rm_1):
                table_data.append({header:val for header, val in zip(header_row, cells)})
        else:
            #if new header, update current header
            header_row = cells
            prev_row_length = len(header_row)
    #combine list of dicts into one larger dict
    combined = {}
    for d in table_data:
        combined.update(d)
    table_data = combined
    return table_data, total_title


def parse_wikitable_achievements(table: BeautifulSoup) -> pd.DataFrame:
    """
    Parses a achievement style wikitable
    Parameters
    ----------
        table: BeautifulSoup
            A beautifulsoup object describing the html of the achievement table
    Returns
    -------
        pd.DataFrame
            A dataframe describing the contents of the achievements table
    """
    header_row = table.find("tr")
    assert isinstance(header_row, Tag)
    headers = [th.get_text(strip=True) for th in header_row.find_all("th") if
               th.get_text(strip = True)]

    rows = []
    for tr in table.find_all("tr")[1:]:
        if tr.get("style") == "display:none":
            continue

        cells = tr.find_all(["td", "th"])
        if not cells:
            continue
        vod_link = None
        cell_values = []
        for c in cells:
            a_tag = c.find("a")
            #checking potential edge cases
            team_span = c.find("span", class_="team-template-image-icon")
            vod_span = c.find("span", class_ = "plainlinks vodlink")
            talent_partners = c.find("div", class_ = "NavContent broadcast-talent-partner-list")

            #parse team
            if team_span and a_tag and a_tag.has_attr("href"):
                team_name = a_tag.get("title", a_tag.get_text(strip=True))
                team_href = a_tag["href"]
                cell_values.append(f"{team_name} ({team_href})")
            elif vod_span and a_tag and a_tag.has_attr("href"):
                vod_link = a_tag['href']
                cell_values.append(vod_link)
            elif talent_partners:
                all_partners = [partner.get_text(strip = True)
                                for partner in talent_partners.find_all("li")]
                if len(all_partners) == 1:
                    all_partners = all_partners[0]
                cell_values.append(all_partners)
            else:
                text = c.get_text(" ", strip=True)
                if len(text) > 0:
                    cell_values.append(text)

        if len(cell_values) > len(headers):#for matches, merging the opponent and result section
            result_idx = headers.index("Result")
            cell_values[result_idx] = f"{cell_values[result_idx]} vs {cell_values[result_idx + 1]}"
            cell_values.pop(result_idx + 1)
        if len(cell_values) < len(headers) and "VOD(s)" in headers:
            #in case no vod, add at the end to make sure same number of cell values and headers
            #this edge case resolution is reall stringent TODO relook
            cell_values.append(vod_link)
        if len(cell_values) == len(headers):
            rows.append(dict(zip(headers, cell_values)))

    df_results = pd.DataFrame(rows)
    return df_results


def parse_team_history(soup_str : BeautifulSoup) -> List[Dict[str, str]]:
    """
    Parses the team history from html
    Parameters
    ----------
        soup_str: BeautifulSoup
            A beautifulsoup object describing the html of the team's infobox
    Returns
    -------
        List[Dict[str, str]]
            A list of dictionaries describing an entry in a team's history
        
    """
    history_div = soup_str.find("div", string="History")
    history_div = history_div.find_next("div", class_="infobox-center") if history_div else None
    history_entries = []
    current_game = None
    if isinstance(history_div, Tag):
        for element in history_div.children:
            if isinstance(element, Tag) and element.name == "b":
                game_name = element.get_text(strip=True)
                current_game = game_name

            if (isinstance(element, Tag)
                and element.name == "div"
                and element.has_attr("style")
                and "font-size:100%" in element["style"]):
                date_div = element.find(
                        "div",
                        style=lambda v: isinstance(v, str) and "float:left" in v
                    )
                dates = date_div.get_text(strip=True) if date_div else None

                team_div = element.find(
                        "div",
                        style=lambda v: isinstance(v, str) and "float:right" in v
                    )
                team_name, team_href = None, None
                team_link = team_div.find("a") if team_div else None
                if isinstance(team_link, Tag):
                    team_name = team_link.get_text(strip=True)
                    team_href = team_link.get("href") if team_link else None

                note_tag = team_div.find("i") if team_div else None
                note = None
                dates = re.findall(r'[\d?]{4}-[\d?]{2}-[\d?]{2}', dates) if dates else [None, None]
                if isinstance(note_tag, Tag):
                    note = note_tag.get_text(strip=True)
                history_entries.append({
                    "game": current_game,
                    "start_date": dates[0],
                    "end_date": dates[1],
                    "team": team_name,
                    "link": team_href,
                    "note": note
                })
    return history_entries

def build_tab_map(data : BeautifulSoup) -> Dict[str, BeautifulSoup]:
    """
    Builds a mapping between tab headers and the contents
    """
    year_map = {}
    for li in data.select(".nav-tabs li[class^='tab']"):
        year_text = li.get_text(strip=True)
        data_count = li.get("class")
        if not data_count:
            continue
        data_count = data_count[0]
        if year_text != "Show All":
            num = re.findall(r'\d+', data_count)
            num = "".join(map(str, num))
            if num in year_map:
                break #found a nested tab, just break out
            year_map[num] = year_text
    year_div_map = {}
    for div in data.select(".tabs-content > div[class^='content']"):
        data_count = div.get("class")
        if not data_count:
            continue
        data_count = data_count[0]
        num = re.findall(r'\d+', data_count)
        num = "".join(map(str, num))
        year = year_map.get(num)
        if year in year_div_map:
            break
        if year:
            year_div_map[year] = div
    return year_div_map

def parse_single_tab_history(tab : Tag, year : Optional[str] = None) -> List[dict[str, str]]:
    """Parses information for a single tab of a team's history"""
    output = []
    entries = tab.find_all("li")
    for entry in entries:
        entry_dict = {}
        date, text = entry.get_text().split(" - ")
        entry_dict['date'] = date
        entry_dict['text'] = text
        entry_dict['year'] = year if year else entry.find_previous(
            ["h1", "h2", "h3", "h4", "h5", "h6"]).get_text(strip = True)
        output.append(entry_dict)
    return output


def parse_wikitable_players(table : Tag) -> pd.DataFrame:
    """Parses information for a wikitable player table"""
    player_rows = table.find_all("tr", class_ = "Player")
    data = []
    for player_row in player_rows:
        player_dict = {}
        for element in player_row.find_all("td"):
            element_class = element.get("class")[0] if element.get("class") else None
            if not element_class:
                continue
            text = element.get_text(strip = True)
            if element_class == "Date":
                element_class, text = text.split(":")[0:2]
            if element_class:
                player_dict[element_class] = re.sub(r"\[[^\]]*\]", "", text)#remove citations
        data.append(player_dict)

    return pd.DataFrame(data)

def parse_wikitable_standins(table : Tag) -> pd.DataFrame:
    """Parses a wikitable describing a team's stand ins"""
    header_row = table.find_all("tr")[1]
    headers = [th.get_text(strip=True) for th in header_row.find_all("th")
               if th.get_text(strip = True)]
    headers[3] = f"{headers[3]}_{headers[2]}"
    headers[4] = f"{headers[4]}_{headers[2]}"
    headers.pop(2)
    rows = []
    for tr in table.find_all("tr")[2:]:
        player_data = []
        for element in tr.find_all("td"):
            if element.find("span", class_ = "flag"):
                continue
            hyperlinks = element.find("a")

            if hyperlinks:
                text = hyperlinks['title']
            else:
                text = element.get_text(strip = True)
            if text == "None":
                player_data.append(text)
             #append extra "None" to keep columns in check if not replacing anyone
            if len(text) > 0:
                player_data.append(text)

        rows.append(dict(zip(headers, player_data)))
    df_results = pd.DataFrame(rows)
    return df_results

def parse_players_raw(text : BeautifulSoup, game :str) -> List[pd.DataFrame]:
    """
    Parses html describing all players for an organization

    Parameters
    ----------
        text: BeautifulSoup
            A beautifulsoup object describing the html of the team's players section
        game: str
            The game being played
    Returns
    -------
        List[pd.DataFrame]
            A list of dataframes describing all tables in the players section
        
    """
    all_players = []
    text_classes = text.get("class")
    if not text_classes:
        return all_players
    if 'table-responsive' in text_classes:
        tables = [text]
    else: tables = text.find_all("div", class_ = "table-responsive")
    for table in tables:
        if "notable temporary stand-ins" in table.get_text().lower():
            player_df = parse_wikitable_standins(table)
            player_df['status'] = 'standin'
        else:
            #print(table)
            player_df = parse_wikitable_players(table)
            header_tag = table.find_previous(["h1", "h2", "h3", "h4", "h5", "h6"])
            player_df['status'] = header_tag.get_text(strip = True) if header_tag else None
        player_df['game'] = game
        all_players.append(player_df)
    return all_players

def parse_achievements(bs4_str : BeautifulSoup
                       ) -> Union[List[pd.DataFrame],
                            List[Dict[str, pd.DataFrame]],
                            pd.DataFrame]:
    """
    Parses html describing the achievements section, used to account for tabbing

    Parameters
    ----------
        bs4_str: BeautifulSoup
            The bs4 object describing the raw html of the achievements table(s)
    Returns
    -------
        Union[List[pd.DataFrame],
            List[Dict[str, pd.DataFrame]],
            pd.DataFrame]
            Either a list of dataframes, dictionaries mapping 
            headers to dataframes or just a dataframe describing the achievements section
        
    """
    all_data = []
    achievements_section = get_all_under_header(bs4_str, "Achievements")
    for section in achievements_section:
        if section.get("class") and any("table" in c for c in section.get("class")):
            table = parse_wikitable_achievements(section)
            all_data.append(table)
        if section.get("class") and  any("tabs" in c for c in section.get("class")):
            table = build_tab_map(section)
            table = {k: parse_wikitable_achievements(v) for k,v in table.items()}
            all_data.append(table)
    if len(all_data) == 1:
        return all_data[0]
    if len(all_data) == 0:
        raise parse_liquipedia_wc.SectionNotFoundException("Could not find achievements section")
    return all_data



"""Initializes ggpyparser package"""
from .parse_liquipedia import parse_liquipedia_html, parse_liquipedia_wc, parse_multiple_liquipedia_pages
from .liquipedia_objects import liquipedia_page, player, team, tournament

__all__ = [
    "parse_liquipedia_html",
    "parse_liquipedia_wc",
    "parse_multiple_liquipedia_pages",
    "liquipedia_page",
    "player",
    "team",
    "tournament"
]



"""Testing the Tournament class"""
import unittest
from pathlib import Path
import pandas as pd
from ggpyscraper.liquipedia_objects import tournament
from ggpyscraper import parse_multiple_liquipedia_pages

GAME = "counterstrike"
TOURNAMENTS = [
    "BLAST/Major/2025/Austin/Stage_3",
    "FISSURE/Playground/1",
    "BLAST/Major/2025/Austin/Playoffs",
    "Intel_Extreme_Masters/2025/Dallas",
    "BLAST/Bounty/2025/Spring/Qualifier",
    "CCT/Season_3/Oceania/Series_1",
    "Hero_Esports/Asian_Champions_League/2025",
    "Intel_Extreme_Masters/2025/Dallas/North_America"
]
TOURNAMENT_TYPES = [
    "swiss", "group_playoff", "playoff", "dbl_playoff",
    "blast_bounty", "lower_tier",  "straight_to_final",
]

PRIZE_TYPES = ["cash", "cash_qual", "non-usd", "weird_blast", "qual"]
PRIZE_TOURNAMENTS = ["FISSURE/Playground/1", "AGES/GO_GAME_Festival/2025",
                     "Hero_Esports/Asian_Champions_League/2025", 
                     "BLAST/Bounty/2025/Spring", 
                     "Intel_Extreme_Masters/2025/Dallas/North_America"]
INFO_TOURNAMENTS = [TOURNAMENTS[1]]

PARTICIPANTS_TOURNAMENTS = ["Hero_Esports/Asian_Champions_League/2025", "FISSURE/Playground/1"]
TALENT_TOURNAMENTS = ["FISSURE/Playground/1", "AGES/GO_GAME_Festival/2025"]

def convert_to_dict(tournaments, values):
    """Convers two lists into a dictionary"""
    return dict(zip(tournaments, values))

PRIZE_SIZES_wc = convert_to_dict(PRIZE_TOURNAMENTS, [(6, 3), (4, 3), (8,19), (4,3), (1,4)])
PRIZE_SIZES_html = convert_to_dict(PRIZE_TOURNAMENTS, [(6, 3), (4, 4), (6,4), (4,4), (4,3)])
MATCH_SIZES = convert_to_dict(TOURNAMENTS, [59, 83, 21, 89, 72, 74, 43, 18])
#wikicode parser will get all map forfeits while html will only get one map
INFO_SIZES_wc = convert_to_dict(INFO_TOURNAMENTS, [37])
INFO_SIZES_html = convert_to_dict(INFO_TOURNAMENTS, [12])
PARTICIPANTS_SIZES = convert_to_dict(PARTICIPANTS_TOURNAMENTS, [12, 16])
TALENT_SIZES = convert_to_dict(TALENT_TOURNAMENTS, [29, 1])
BASE_PATH = Path("tests/assets/tournaments")


class TestTournaments(unittest.TestCase):
    """Test the Tournament class"""

    @classmethod
    def setUpClass(cls):
        # Build wikicode-backed Tournament objects (one network call batching page names)
        cls.all_tournaments_wikicode = parse_multiple_liquipedia_pages.create_multiple_pages(
            game=GAME, page_names=TOURNAMENTS, page_ts="tournament"
        )

        # Build HTML-backed Tournament objects from local fixtures
        cls.all_tournaments_html = {}
        for t_name in TOURNAMENTS:
            path = BASE_PATH / f"{t_name.replace('/', '_')}.txt"
            with path.open("r", encoding="utf-8") as f:
                raw_html = f.read()
            t_obj = tournament.Tournament.from_raw_str(
                name=t_name, game=GAME, action="html", response=raw_html
            )
            cls.all_tournaments_html[t_name] = t_obj
    def setup_test_tournaments(self, tournaments,
                               ground_truths, method_name, mode="wikicode", both = False):
        """Sets up and runs tests for tournaments."""
        tourn_dict = (self.all_tournaments_wikicode if
                       mode == "wikicode" else self.all_tournaments_html)
        results_tournament = {name: t_obj for name, t_obj in
                            tourn_dict.items() if
                            name in tournaments}
        for to_name, to_obj in results_tournament.items():
            with self.subTest(name=to_name, mode=mode):
                method = getattr(to_obj, method_name)
                results = method()
                if isinstance(results, list):
                    results = results[0]
                if isinstance(results, dict):
                    results = len(results)
                if isinstance(results, pd.DataFrame):
                    results = results.shape if both else results.shape[0]
                expected_size = ground_truths[to_name]
                self.assertEqual(
                    results,
                    expected_size,
                    msg=f"{to_name}, {mode}",
                )
    def test_results_wc(self):
        """Validate match counts using WC parse."""
        self.setup_test_tournaments(TOURNAMENTS, MATCH_SIZES,
                                    "get_results", mode="wikicode")

    def test_results_html(self):
        """Validate match counts using HTML parse."""
        self.setup_test_tournaments(TOURNAMENTS,  MATCH_SIZES,
                                     "get_results", mode="html")
    def test_prizes_html(self):
        """Validate prize counts using HTML parse."""
        self.setup_test_tournaments(PRIZE_TOURNAMENTS, PRIZE_SIZES_html,
                                    "get_prizes",  mode="html", both = True)
    def test_prizes_wc(self):
        """Validate prize counts using WC parse."""
        self.setup_test_tournaments(PRIZE_TOURNAMENTS, PRIZE_SIZES_wc,
                                    "get_prizes",  mode="wikicode",both = True)

    def test_info_wc(self):
        """Validate info counts using WC parse."""
        self.setup_test_tournaments(INFO_TOURNAMENTS, INFO_SIZES_wc,
                                    "get_info",  mode="wikicode")
    def test_info_html(self):
        """Validate info counts using HTML parse."""
        self.setup_test_tournaments(INFO_TOURNAMENTS, INFO_SIZES_html,
                                    "get_info",  mode="html")

    def test_participants_wc(self):
        """Validate participant counts using WC parse."""
        self.setup_test_tournaments(PARTICIPANTS_TOURNAMENTS, PARTICIPANTS_SIZES,
                                    "get_participants",  mode="wikicode")
    def test_participants_html(self):
        """Validate participant counts using HTML parse."""
        self.setup_test_tournaments(PARTICIPANTS_TOURNAMENTS, PARTICIPANTS_SIZES,
                                    "get_participants",  mode="html")

    def test_talent_wc(self):
        """Validate participant counts using WC parse."""
        self.setup_test_tournaments(TALENT_TOURNAMENTS, TALENT_SIZES,
                                    "get_talent",  mode="wikicode")
    def test_talent_html(self):
        """Validate participant counts using HTML parse."""
        self.setup_test_tournaments(TALENT_TOURNAMENTS, TALENT_SIZES,
                                    "get_talent",  mode="html")


if __name__ == "__main__":
    unittest.main()



"""
Helper Module describing functions to parse data with a wikicode-style input from Liquipedia as well
as additional exceptions and request functions

Dependencies
------------
- BeautifulSoup
- pandas
- regex
- numpy
- requests
- mwparserfromhell: MediaWiki wikitext parsing

Raises
------
SectionNotFoundException
    Raised when a section like achievements or gear is not found or supported using a given method.
CouldNotReadJsonException   
    Raised when the json received from the Liquipedia API is invalid

"""
from typing import Callable, Dict, List, Union
import re
import mwparserfromhell as mw
import pandas as pd
import requests


class SectionNotFoundException(Exception):
    """
    Handles scenarios where the section could not be found on the page
    """
class CouldNotReadJsonException(Exception):
    """
    Handles a json which could not be read
    """
def make_request(user: str, game: str, page_name: str, action: str) -> Dict[str, str]:
    """
    Makes a call to the Liquipedia API

    Parameters
    ----------
        user: str
            The string describing the user as requested by Liquiepdia ToS
        game: str 
            The game of the page being parsed
        page_name: str
            The name of the page being parsed, essentially liquipedia/game/(x)
        action: str 
            Whether to use html or wikicode parsing, "query" means wikicode, "parse" means html

    Returns
    -------
        page_to_str: dict(str, str)
            A mapping between the page name and the result of the request

    """
    action_map = {'wikicode': 'query', 'html': 'parse'}
    api_action = action_map.get(action)
    headers = {
            "User-Agent": user,
            "Accept-Encoding": "gzip"
        }

    try:
        request_params={
                "action": api_action,
                "format": "json"
            }
        if api_action == 'parse':
            request_params['page'] = page_name
        else:
            request_params["rvprop"] =  "content"
            request_params['rvslots'] = 'main'
            request_params['titles'] = page_name
            request_params['prop'] = 'revisions'
        response = requests.get(
            f"https://liquipedia.net/{game}/api.php",
            headers=headers,
            params = request_params,
            timeout=10
        )
        response.raise_for_status()
        try:
            if action  == 'wikicode':
                response = response.json()['query']['pages']
                page_to_str = {}
                for page in response.values():
                    title = page['title']
                    raw_str = page['revisions'][0]['slots']['main']['*']
                    page_to_str[title.lower().strip().replace(" ", "_")] = raw_str

            else:
                response = response.json()['parse']
                title = response['title']
                raw_str = response['text']['*']
                page_to_str = {title.lower().strip().replace(" ", "_"): raw_str}
            return page_to_str

        except KeyError as e:
            raise CouldNotReadJsonException(f"Could not \
                Read JSON Request Result, indicating potential input string issues: {e}") from e
    except requests.exceptions.Timeout as e:
        raise TimeoutError("Request to Liquipedia API timed out.") from e
    except requests.exceptions.RequestException as e:
        raise ConnectionError(f"Request to Liquipedia API failed: {e}") from e
    except Exception as e:
        raise RuntimeError(f"Unexpected error in _make_request: {e}") from e

def parse_player_team_history(raw_str: str) -> pd.DataFrame:
    """
    Parses the team history for a player in their infobox

    Parameters
    ----------
        raw_str: str
            A string describing the wikicode of the player's infobox

    Returns
    -------
        pd.DataFrame
            A Data Frame describing the team history of the given player

    """
    #regex which finds bolded parts and grabs the bold and the immediate expression
    pattern = r"'''(.*?)'''\s*(.*?)(?=(?:'''|$))"


    matches = re.findall(pattern, raw_str, flags=re.DOTALL)
    pattern = r"{{TH|(.*?)}}"
    game_teams = []
    for game, text in matches:
        pattern = r"\{\{TH\|.*?\}\}"
        #parse each team
        teams = re.findall(pattern, text, flags = re.DOTALL)
        for team in teams:
            team_dict = {}
            pattern = r"\{\{TH\|([^|]+?)\|([^|}]+)(?:\|([^}]+))?\}\}"
            match = re.match(pattern, team)
            if match:
                date_range = match.group(1).strip()
                team_dict["team"] = match.group(2).strip()
                team_dict['start'], team_dict['end'] = re.split(r"\s*[−–—]\s*",
                                                                date_range, maxsplit=1)
                team_dict['status'] = match.group(3).strip() if match.group(3) else None
                team_dict['game'] = game
                game_teams.append(team_dict)
    return pd.DataFrame(game_teams)
def parse_team_name(name : str) -> str:
    """Parses the team name from a wikicode match"""
    return re.search(r"\{\{TeamOpponent\|([^}|]*)", name)[1] # type: ignore


def parse_series_data(raw_str: str, regex: str,
                      cleaning_function: Callable = lambda x: x) -> Dict[str, str]:
    """
    Utility function used to find patterns with a regex and apply a cleaning function to a string

    Parameters
    ----------
        raw_str: str
            A string describing a wikicode 
        regex: str
            The regular expression used to build a 2 length tuple depending on the pattern
        cleaning_function: Callable
            A function to apply to the second value of the tuple

    Returns
    -------
        Dict[str, str]
            A dictionary mapping each parts of the tuple to one another
    """
    parsed =re.findall(regex, str(raw_str), re.DOTALL)
    return {key: cleaning_function(value) for key, value in parsed}

def parse_series(series_info: str) -> pd.DataFrame:
    """
    Parses a series of matches

    Parameters
    ----------
        series_info: str
            A string describing a wikicode of a series
        game: str
            The video game played during the series

    Returns
    -------
        pd.DataFrame
            A dataframe describing information about each map in the series
    """
    #get games(single maps) -
    matches = extract_maps(series_info)

    matches = pd.DataFrame(matches)
    #get teams:
    pattern = r"(opponent\d+)\s*=\s*(\{\{TeamOpponent\|.*?\}\})"
    team_names = parse_series_data(series_info, pattern, parse_team_name)
    if len(team_names) == 0:
        return matches
    #get date:
    matches[['opponent_1', 'opponent_2']] = team_names['opponent1'], team_names['opponent2']
    parsed = mw.parse(series_info)
    match_tpl = next(t for t in parsed.filter_templates(recursive=True)
                 if t.name.matches("Match"))

    for p in match_tpl.params:
        if not ("map" in p.name or "opponent" in p.name):
            matches[str(p.name)] = str(p.value).strip()
    return matches


def parse_grouped_games(name:str, info:List[str]) -> List[pd.DataFrame]:
    """
    Parses games in a group stage setting

    Parameters
    ----------
        name: str
            The name of the stage(e.g. "Group A")
        info: List[str]
            A list of strings describing the wikicode of the grouped games
    Returns
    -------
        List[pd.DataFrame]
            A list of dataframes describing all of the series in the groups
    """
    alldfs = []
    if isinstance(info, str):
        info = mw.parse(info)
    for template in info.filter_templates(recursive=True):
        if template.name.matches("Matchlist") or template.name.matches("SingleMatch"):
            series_idx = 0
            for subtemplate in template.params:
                if "title=" in str(subtemplate):
                    name = subtemplate.split("=")[1]
                if "{{Match" in subtemplate or "{{SingleMatch" in subtemplate:
                    match_df = parse_series(str(subtemplate))
                    match_df['stage'] = name
                    match_df['series_id'] = series_idx
                    series_idx += 1
                    alldfs.append(match_df)
    return alldfs

def parse_bracket(info: str) -> List[pd.DataFrame]:
    """
    Parses games in a bracket stage setting

    Parameters
    ----------
        info: string
            A string describing wikicode
        game: str
            The game being played

    Returns
    -------
        List[pd.DataFrame]
            A list of dataframes describing all of the series in the bracket
    """
    alldfs = []
    series_idx = 0

    #first try to get stage name from the RxMxheader
    regex = (
        r"(\|R\d+M\d+header=[^\n]+)"  #turn into tuple of (|RxMxheader=name, text)
        r"(.*?)"                          
        r"(?=\|R\d+M\d+header=|\Z)"       
    )
    #if header declarations are at top, just takes last RxMx
    # if uses a mixture of two heading styles(header= and <-->, does not work
    # see dota2/Esports_World_Cup/2025/North_America
    if len(re.findall(regex, str(info), re.DOTALL)) > 0:
        #if we're not using comments to parse headers, remove all comments
        info = re.sub(r"<!--.*?-->", "", str(info), flags=re.DOTALL)
        header_regex = r"\|R(\d+)M\d+header=([^\n]+)"
        #remove comments
        #headers_list = re.findall(header_regex, str(info))
        headers = dict(re.findall(header_regex, str(info)))
        if len(headers) > 1:
            #should have multiple headers,
            for round_num, stage_name in headers.items():
                # Find all matches for this round
                regex = (
                        r"\|R" + str(round_num) +
                        r"M\d+=\s*(\{\{Match.*?\}\})\s*(?=\|R" +
                        str(round_num) + r"M\d+=|\|R\d+M\d+=|\Z)"
                    )
                round_matches = re.findall(regex, str(info), re.DOTALL)
                for match_text in round_matches:
                    match_df = parse_series(match_text)
                    match_df['stage'] = stage_name
                    alldfs.append(match_df)
        else: #if not, parse manually up to down - dealing
            #with issues where two headers are marked R1 but in different places
            stages = parse_series_data(info, regex)
            for stage, text in stages.items():
                stage = stage.split("=")[1]
                matches = re.split(r'\|R\d+M\d+=', text)
                for match in matches:
                    if "{{Match" in match:
                        match_df = parse_series(match)
                        match_df['stage'] = stage
                        match_df['series_id'] = series_idx
                        series_idx += 1
                        alldfs.append(match_df)
    else:

        #if fails, look at <!--stage-->
        #can be situations where <!--x--> is being used as a temporary value for
        # an event that has not happened yet - i have hardcoded removing these comments
        #but this is not a good solution
        regex =  r'<!--\s*(.*?)\s*-->\s*(.*?)(?=\s*<!--\s*\w+|$)'
        stages = parse_series_data(remove_non_stage_comments(str(info)), regex)
        if len(stages) == 0:#no headers found
            #maybe inferring headers from Bracket/4L2DS?
            stages = {"bracket": str(info)}
        for stage, text in stages.items():
            matches = [t for t in mw.parse(text).filter_templates(recursive=True)
                    if t.name.matches("Match")]
            for match in matches:
                match_df = parse_series(match)
                match_df['stage'] = stage
                match_df['series_id'] = series_idx
                series_idx += 1
                alldfs.append(match_df)
    return alldfs
def parse_games(stage, info):
    """
    Determines whether a bracket or grouped based parser should be used

    Parameters
    ----------
        stage: str
            A string describing the name of the stage
        info: List[str]
            A list of strings describing wikicode for a page
        game: str
            The game being played

    Returns
    -------
        List[pd.DataFrame]
            A list of dataframes describing all the matches found within info
    """
    if "Bracket" in info:#might be worth doing an any check here
        new_games =  pd.concat(parse_bracket(info))
    else:
        new_games =  pd.concat(parse_grouped_games(stage, info))
    return new_games

def parse_team(text:str) -> pd.Series:
    """Extracts team information from wikicode text from the participants section"""
    #get name
    match = re.search(r"\|\s*team\s*=\s*([^|}]+)", text)
    team = match.group(1).strip() if match else None

    #get qualification method
    match = re.search(r"\|\s*qualifier\s*=\s*([^|}]+)", text)
    qualifier =  match.group(1).strip("/[") if match else None

    match = re.findall(r"\b(p\d+|c)\s*=\s*([^\s|}]+)", text)
    players = {k: v for k, v in match} if match else None
    #find dnps
    match = re.search(r"\b(xxdnp)\s*=\s*(true)\b", text)
    dnps = (value for key, value in  match.groups()) if match else ()

    team_dict = {"team": team, "qualifier": qualifier, "dnps": dnps}
    if team_dict and players:
        team_dict.update(players)
    return pd.Series(team_dict)

def parse_talent(text:str) -> pd.DataFrame:
    """Parses information about a talent member"""

    #get broadcast role
    match = re.search(r"\|\s*position\s*=\s*([^|}]+)", text)
    position = match.group(1).strip() if match else None

    #get broadcast language
    match = re.search(r"\|\s*lang\s*=\s*([^|}]+)", text)
    language =  match.group(1).strip() if match else None

    match = re.findall(r"\|b\d+\s*=\s*([^\|}]+)", text)
    names = [m.strip() for m in match]

    return pd.DataFrame(data = {
        "name": names,
        "language": [language] * len(names),
        "position": [position] * len(names)
    })
def parse_prizes(text: mw.parser.Parser,
                 match_1 = "slot",
                 match_2 = r"(?i)prizepool") -> pd.DataFrame:
    """Parses prize information for a tournament

    Parameters
    ----------
        text: mw.parser.Parser
            The parsed containing the prizes to parse
        match_1 : str
            The outer template to match
        match_2 : str
            The inner template to match
    Returns
    -------
        pd.DataFrame
            A dataframe describing the prize pool
    """

    wikicode = text
    s_m = {}
    slots = [t for t in wikicode.filter_templates()
         if t.name.strip().lower() == match_1]
    slot_meta_tpls = wikicode.filter_templates(matches=match_2, recursive=False)
    if slot_meta_tpls:
        slot_meta = slot_meta_tpls[0]
        s_m = {str(p.name): str(p.value) for p in slot_meta.params if
               not str(p.name).strip().isdigit()}
    all_prizes = []
    for slot in slots:
        slot_data = {}
        for param in slot.params:
            val = str(param.value)
            slot_data[str(param.name)] = val
        slot_data.update(s_m)
        all_prizes.append(slot_data)
    if len(all_prizes) == 0:
        raise SectionNotFoundException("No prize information found")
    all_prizes = pd.DataFrame(all_prizes)
    if "count" in all_prizes.columns:
        all_prizes['count'] = all_prizes['count'].fillna(1)
    return all_prizes


def get_name_content_map(text: str) -> Dict[str, str]:
    """Builds a mapping between name(x) and content(x) as described in wikicode"""
    pattern = r"\|name(\d+)=(.*)"
    key_mapping = dict(re.findall(pattern, text))
    #print(text)
    pattern = r"\|content(\d+)=\s*\n(.*?)(?=\n\|content\d+=|\n\|name\d+=|\n\}\})"
    values = dict(re.findall(pattern, text, flags=re.S))
    if len(values) == 0:
        pattern = re.compile(r"""
        \{\{\s*tabs\s+dynamic/tab\s*\|\s*(\d+)\s*\}\}   
        \s*(?:<!--.*?-->)?\s*                          
        (.*?)                                          
        (?=                                            
            \{\{\s*tabs\s+dynamic/tab\s*\|\s*\d+\s*\}\}
        | \{\{\s*tabs\s+dynamic/end\s*\}\}          
        )
        """, re.IGNORECASE | re.DOTALL | re.VERBOSE)
        values = dict(pattern.findall(text))
    mapping = {key_mapping[k]: v for k,v in values.items()}
    return mapping

def parse_news_str(raw_str: str) -> Dict[str, Union[str, List[str]]]:
    """Parses a string describing the news"""
    pattern = r"<ref.*?>(.*?)</ref>"
    cleaned = re.sub(r"\{\{\s*flag\b[^{}]*\}\}", "", raw_str, flags=re.IGNORECASE)
    ref_content = re.findall(pattern, str(cleaned), flags=re.S)

    refs = []
    for ref in ref_content:
        pairs = re.findall(r"(\w+)=([^|}]+)", ref)
        ref_dict = {k.strip(): v.strip("[ ]") for k, v in pairs}
        refs.append(ref_dict)

    entry = re.sub(r"\[\[([^\]]+)\]\]", r"\1", str(raw_str))#remove [[]]
    entry = re.sub(r"<ref.*?>(.*?)</ref>", "", entry)#remove reference
    entry = re.sub(r"(\*|')", "", entry).strip()
    entry = re.split(r"\s-\s", entry)
    if len(entry)  == 2:
        return {"date":entry[0], "text": entry[1], "references":refs}
    return {}

def get_lowest_subsections(section: mw.wikicode.Wikicode) -> List[mw.wikicode.Wikicode]:
    """
    Recursively gets all lowest-level subsections for a section.
    """
    subsections = section.get_sections(include_lead=False, include_headings=True)[1:]
    if not subsections:
        return [section]

    lowest = []
    for sub in subsections:
        lowest.extend(get_lowest_subsections(sub))
    return lowest#rm duplicates

def parse_person(text: str) -> Dict[str, str]:
    """
    Parses information about a person given the wikicode string
    """
    person = re.sub(r"<ref.*?>.*?</ref>|<ref.*?/>", "", text) #remove reference

    pattern = r'(\w+)\s*=\s*(.*?)(?=\|\w+\s*=|$)'
    pairs = re.findall(pattern, person)
    person_dict = {k.strip(): v.strip() for k, v in pairs}

    joindate = person_dict['joindate'].split("|") if 'joindate' in person_dict else None
    if joindate and len(joindate) > 1:
        person_dict['joindate'] = joindate[1]
        person_dict['joindate_note'] = joindate[2]

    if 'tournament' in person_dict:
        tournaments = re.findall(r'\[\[[^\]|]+\|([^\]]+)\]\]', person)
        person_dict['tournament'] = tournaments
    return person_dict



def remove_non_stage_comments(s: str) -> str:
    """
    Removes all comments that do not contain stage information
    """
    #regex for comments and stage names
    comment = re.compile(r"<!--\s*(?P<txt>(?:(?!-->).)*?)\s*-->", re.DOTALL)
    #I hate this hardcoding but idk a better way rn
    stages = re.compile(r"\b(round|stage|quarterfinals?|semifinals?|final|match|group)\b",
                         re.IGNORECASE)
    def keep_only_stage(m: re.Match) -> str:
        #for all comments, check if it contains a stage word
        inner = m.group("txt")
        return m.group(0) if stages.search(inner) else ""
    return comment.sub(keep_only_stage, s)
def extract_maps(wikitext):
    """
    Extracts maps from the wikitext
    """
    code = mw.parse(wikitext)
    out = []
    for t in code.filter_templates(recursive=True):
        if t.name.strip().lower() == "map":
            params = {str(p.name).strip(): str(p.value).strip() for p in t.params}
            out.append(params)

    return out



"""
Module for representing and extracting information about esports players from Liquipedia.

Defines the Player class, inherits from LiquipediaPage and provides
methods to fetch and parse player-specific data such as team history, gear setup, and
achievements from Liquipedia. 

Classes
-------
Player
    Represents an individual player and provides methods to extract structured information
    about their competitive history, gear settings, and tournament achievements.

Dependencies
------------
- BeautifulSoup
- pandas
- mwparserfromhell: For parsing MediaWiki markup.
- parse_liquipedia, parse_liquipedia_html: Custom modules for parsing specific data blocks.
- LiquipediaPage: Base class for interacting with Liquipedia.
"""
from typing import Dict, List, Union
import regex as re
from bs4 import BeautifulSoup
import pandas as pd
import mwparserfromhell as mw
from ggpyparser.parse_liquipedia import parse_liquipedia_html, parse_liquipedia_wc
from ggpyparser.liquipedia_objects import liquipedia_page


class Player(liquipedia_page.LiquipediaPage):
    """
    A class to represent a player.


    Methods
    -------
    get_info(infobox_name = "Infobox player"):
        Parses information from the players infobox
    get_gear()
        Parses information about the gear the player uses
    get_achievements()
        Parses information about the player's achievements(html only)
    
    """
    def __init__(self, game :str, name: str,
                 user: str, action : str = "wikicode"
                 ) -> None:
        """
        Creates a player object

        Parameters
        ----------
        game: str
            The game being played
        name: str
            The page name, found by liquipedia.com/game/(name)
        user: str
            The user, as requested by liquipedia ToS
        action: str
            Whether html(action = "parse") or wikicode(action = "wikicode") parsing should occur
        """
        super().__init__(game, name, user=user, action = action)

    def get_info(self, infobox_name: str = "Infobox player"
    ) -> Dict[str, Union[str, pd.DataFrame, List[Dict[str, str]]]]:
        """
        Gets the information from a player's infobox

        Parameters
        ----------
            infobox_name: str
                The name of the infobox to parse, defaults to "infobox player", 
                custom to each type of object
        
        Returns
        -------
            Dict[str, str]
                A dictionary describing the contents of the infobox
        """
        if self.action == "wikicode":
            return self._get_info_wc(infobox_name = infobox_name)
        return self._get_info_html()

    def _get_info_html(self) -> Dict[str, Union[str, pd.DataFrame, List[Dict[str, str]]]]:
        """
        Private method to get infobox data using an html parser

        Returns
        -------
            Dict[str, str]
                A dictionary describing the contents of the infobox
        """
        info_dict = super()._get_info_html()

        #parse previous teams
        souped = BeautifulSoup(self.get_raw_str(), "html.parser")
        history_entries = parse_liquipedia_html.parse_team_history(souped)
        team_history = pd.DataFrame(history_entries)
        team_history['game'] = team_history['game'].fillna(self.game)
        info_dict['team_history'] = team_history
        return info_dict

    def _get_info_wc(self, infobox_name : str ="Infobox player"
                     ) -> Dict[str, Union[str, pd.DataFrame, List[Dict[str, str]]]]:
        """
        Private method to get infobox data using a wikicode parser

        Parameters
        ----------
            infobox_name: str
                The name of the infobox to parse, defaults to "infobox player", 
                custom to each type of object

        Returns
        -------
            Dict[str, str]
                A dictionary describing the contents of the infobox
        """
        info_dict =  super()._get_info_wc(infobox_name)
        if 'team_history' in info_dict and isinstance(info_dict['team_history'], str):
            info_dict['team_history'] = parse_liquipedia_wc.parse_player_team_history(
                info_dict['team_history'])
        else:
            info_dict['team_history'] = pd.DataFrame()
        return info_dict

    def get_gear(self) -> Dict[str, pd.DataFrame]:
        """
        Gets information about the player's gear

        Returns
        -------
            Dict[str, pd.DataFrame]
                A dictionary mapping gear to the contents of each gear section
        """
        if self.action == "wikicode":
            return self._get_gear_wc()
        return self._get_gear_html()

    def _get_gear_html(self):
        """
        Private method to parse gear information with html

        Returns
        -------
            Dict[str, pd.DataFrame]
                A dictionary mapping gear to the contents of each gear section
        """
        all_data = {}
        souped = BeautifulSoup(self.get_raw_str(), "html.parser")
        sections = parse_liquipedia_html.get_all_under_header(souped,  "Gear_and_Settings")
        for table in sections:
            if table.get("class") and any("table" in c for c in table.get("class")):
                table, title  = parse_liquipedia_html.parse_wikitable_hdhd(table, rm_1 = True)
                all_data[title] = table
                #print(table)
        if len(all_data) == 0:
            raise parse_liquipedia_wc.SectionNotFoundException("Could not find gear section")
        return all_data

    def _get_gear_wc(self):
        """
        Private method to parse gear information with wikicode

        Returns
        -------
            Dict[str, pd.DataFrame]
                A dictionary mapping gear to the contents of each gear section
        """
        parsed = mw.parse(self.get_raw_str())
        gear_dict = {}
        for section in parsed.get_sections(matches=r"(?i)^gear and settings"):
            for template in section.filter_templates():
                subgear_dict = {}
                heading = template.name.strip().replace(" table", "")
                for param in template.params:
                    #check for references
                    pattern = r"<ref.*?>(.*?)</ref>"
                    ref_content = re.findall(pattern, str(param.value).strip(), flags=re.S)
                    if len(ref_content) > 0:
                        subgear_dict[str(param.name)] = [ref.strip("[ ]") for ref in ref_content]
                    else:
                        subgear_dict[str(param.name)] = str(param.value).strip()
                    
                gear_dict[heading] = subgear_dict
        return gear_dict
    def get_achievements(self) -> Union[List[pd.DataFrame],
                            List[Dict[str, pd.DataFrame]],
                            pd.DataFrame]:
        """
        Parse a player's achievements

        Returns
    -------
        Union[List[pd.DataFrame],
            List[Dict[str, pd.DataFrame]],
            pd.DataFrame]
            Either a list of dataframes, dictionaries mapping 
            headers to dataframes or just a dataframe describing the achievements section
        """
        if self.action == "wikicode":
            raise parse_liquipedia_wc.SectionNotFoundException(
                "Cannot parse achievements section using a wikicode parse, try html parse")
        return parse_liquipedia_html.parse_achievements(BeautifulSoup(
            self.get_raw_str(),"html.parser")).reset_index(drop = True)



"""
Module to parse the general Liquipedia pages

Dependencies
------------
- BeautifulSoup
- pandas
- liquipedia_page: Base class for interacting with Liquipedia.
"""
import pandas as pd
from bs4 import BeautifulSoup
from ggpyparser.liquipedia_objects import liquipedia_page
def parse_collapsable_tables(soup : BeautifulSoup) -> pd.DataFrame:
    """
    Parses a beautifulsoup collapsible table

    Parameters
    ----------
        soup : BeautifulSoup
            A beautifulsoup object to parse
    
    Returns
    -------
        pd.DataFrame:
            A pandas dataframe describing the results of all the tables

    """
    tables = soup.select("table.wikitable.collapsible")
    player_data = []
    for table in tables:
        rows = table.find_all('tr')
        headers = rows[:2]
        large_header = headers[0].get_text()
        subheaders = headers[1].find_all("th")
        header_map = dict(zip(range(len(subheaders)),
                                [h.get_text(strip = True) for h in subheaders]))
        for person in rows[2:]:
            header_idx = 0
            elements = person.find_all("td")
            person_dict = {'type': large_header}
            for element in elements:
                header = header_map[header_idx]
                links = element.find_all("a")
                link_list = []
                for link in links:
                    link_list.append(link.get("href"))
                link_list = list(set(link_list))
                person_dict[header] = element.get_text(strip = True)
                if len(link_list) > 0:
                    if header.lower() == "links":
                        person_dict[header] = link_list
                    else:
                        person_dict[f"{header}_links"] = link_list
                header_idx += 1
                #print(person_dict)
            player_data.append(person_dict)
        #county_players = {"country": country, "name": player.get_text() for player in players}
    return pd.DataFrame(player_data)

def parse_tournaments(name : str, game: str, user: str) -> pd.DataFrame:
    """
    Parses general tournament pages like https://liquipedia.net/counterstrike/S-Tier_Tournaments

    Parameters
    ----------
        name: str
            Name of the page, found from https://liquipedia.net/counterstrike/{page}
        game: str
            The game being played
        user: str
            Information about the current project
    
    Returns
    -------
        pd.DataFrame
            A dataframe describing the contents of the tournament webpage
    """
    raw_str = liquipedia_page.LiquipediaPage(game = game, name = name,
                                             action = "html", user = user).get_raw_str()
    tournament_data = []
    souped = BeautifulSoup(raw_str, "html.parser")
    cell_name = "div"
    rows = souped.find_all("div", class_ = "divRow")
    if len(rows) == 0:
        rows = souped.find_all("div", class_ = "gridRow")
        cell_name = "grid"
    for row in rows:
        cells = row.find_all("div", class_ = f"{cell_name}Cell")
        row_data = {}
        for cell in cells:
            cell_class = cell.get("class")
            text = list(set([val for val in cell.get_text(separator= "|").split("|")
                             if len(val.strip()) > 0]))
            if "Header" in cell_class:
                cell_class.remove("Header")
            hrefs = list(set([href['href'] for href in cell.find_all('a')]))
            header_name = cell_class[-1]
            if "Place" not in header_name and "Qualified" not in header_name:
                #if not teams, then don't keep separated
                text = "".join(text)
            if "EventDetails" in header_name:
                #don't know true header, must infer
                if "Left" in header_name:
                    #this is wayyyy to stringent, rework moving forward
                    header_name = "Date" if "55" in header_name else header_name
                    header_name = "Location" if "60" in header_name else header_name
                if "Right" in header_name:
                    header_name = "Prize" if "45" in header_name else header_name
                    header_name = "PlayerNumber" if "40" in header_name else header_name
            text = text[0] if isinstance(text, list) and len(text) == 1 else text
            row_data[header_name] = text
            if len(hrefs) > 0:
                row_data[f"{header_name}_links"] = hrefs
        tournament_data.append(row_data)
    return pd.DataFrame(tournament_data)

def parse_teams(region: str, game: str, user:str) -> pd.DataFrame:
    """
    Parses general teams pages like https://liquipedia.net/counterstrike/Portal:Teams/Europe

    Parameters
    ----------
        region: str
            The region of interest, found from 
            https://liquipedia.net/counterstrike/Portal:Teams/{region}

        game: str
            The game being played
        user: str
            Information about the current project
    
    Returns
    -------
        pd.DataFrame
            A dataframe describing the contents of the teams webpage
    """
    name = f"Portal:Teams/{region}"
    raw_str = liquipedia_page.LiquipediaPage(game = game, name = name,
                                             action = "html", user = user).get_raw_str()
    soup = BeautifulSoup(str(raw_str), "html.parser")

    active_teams = parse_collapsable_tables(soup)
    active_teams['active'] = True
    #parse inactive teams
    inactive = []
    if game == "counterstrike":
        tables = soup.select("table.wikitable.smwtable")
        for table in tables:
            inactive.append({'type': table.get_text(strip = True), 'active': False,})

    else:
        disbanded = soup.find("span", id = "Disbanded_teams").findNext('div')
        for li in disbanded.find_all('li'):
            inactive.append({'type': li.get_text(strip = True), 'active' : False})
    inactive = pd.DataFrame(inactive)
    return pd.concat([active_teams, inactive]).rename(columns = {"type": "team"})

def parse_players(region: str, game: str, user: str) -> pd.DataFrame:
    """
    Parses general teams pages like https://liquipedia.net/counterstrike/Portal:Players/Europe

    Parameters
    ----------
        region: str
            The region of interest, found from 
            https://liquipedia.net/{game}/Portal:Players/{region}
            N.B. For Africa and the Middle East, the region is "Africa_&_Middle_East" 
            not "Africa_%26_Middle_East"

        game: str
            The game being played
        
        user: str
            Information about the current project
    
    Returns
    -------
        pd.DataFrame
            A dataframe describing the contents of the players webpage
    """
    name = f"Portal:Players/{region}"
    raw_str = liquipedia_page.LiquipediaPage(game = game, name = name,
                                             action = "html", user = user).get_raw_str()
    soup = BeautifulSoup(str(raw_str), "html.parser")
    tables = soup.find_all("table", class_=["wikitable", "collapsible"])
    if game == "counterstrike":
        #cs has a different player style for some reason
        player_dict = []
        for table in tables:
            country = table.find('th').get_text()
            players = table.find_all("td")
            for player in players:
                name = player.get_text().split(" - ")[1]
                tag = player.get_text().split(" - ")[0]
                player_dict.append({"country": country, "name":name, "tag":tag})
            #county_players = {"country": country, "name": player.get_text() for player in players}
        return pd.DataFrame(player_dict)
    else:
        soup = BeautifulSoup(raw_str)
        return parse_collapsable_tables(soup)

def parse_banned_players(game : str, user : str, company: str = None) -> pd.DataFrame:
    """
    Parses banned players page

    Parameters
    ----------
        company: str
            The company banning the players, found from 
            https://liquipedia.net/{game}/Banned_Players/{company}
            If no company, use None
        game: str
            The game being played
    
    Returns
    -------
        pd.DataFrame
            A dataframe describing the contents of the banned players webpage
    """
    name = f"Banned_Players/{company}"
    if company is None:
        name = "Banned_players"
    raw_str = liquipedia_page.LiquipediaPage(game = game, name = name,
                                              action = "html", user = user).get_raw_str()
    soup = BeautifulSoup(str(raw_str), "html.parser")
    tables = soup.find_all("div", class_ = "divTable Ref")
    banned = []
    for table in tables:
        players = table.find_all("div", class_ = "divRow mainpage-transfer-neutral")
        for player in players:
            banned_dict = {}
            name = player.find("div", class_ = "divCell Name")
            banned_dict['name'] = name.get_text() if name else None

            team = player.find("div", class_ = "divCell Team")
            banned_dict["team"] = team.find("a").get("title") if team and team.find("a") else None

            for div in player.find_all("div", class_ = "divCell"):
                #this is really dumb but idk other ways to find reason besides parsing all divCells
                if div.get("class") == ['divCell']:
                    banned_dict['reason'] = div.get_text()

            start, end = player.find_all("div", class_ = "divCell Date")
            banned_dict['start'] = start.get_text() if start else None
            banned_dict['end'] = end.get_text() if end else None
            banned.append(banned_dict)

    return pd.DataFrame(banned)

def parse_transfers(name:str, game:str, user: str) -> pd.DataFrame:
    """
    Parses transfers page

    Parameters
    ----------
        name: str
            The time period considered, found from
            https://liquipedia.net/counterstrike/{name}
            Usually "Transfers/{time}" or "Player_Transfers/{time}"

        game: str
            The game being played
        user: str
            Information about the current project

    
    Returns
    -------
        pd.DataFrame
            A dataframe describing the contents of transfers page
    """
    raw_str = liquipedia_page.LiquipediaPage(game = game, name = name, 
                                             action = "html", user = user).get_raw_str()
    soup = BeautifulSoup(str(raw_str), "html.parser")
    tables = soup.find_all("div", class_ = "divTable mainpage-transfer Ref")
    transfers_list = []

    for table in tables:
        transfers  = table.find_all("div", class_ = "divRow")
        for transfer in transfers:
            transfer_dict = {}
            date = transfer.find("div", class_ = "divCell Date")
            transfer_dict['date'] = date.get_text()

            # Extract player names
            transfer_dict['names'] = [
            block.select_one(".name a").get_text(strip=True)
            for block in transfer.select("div.block-player")
            ]
            transfer_dict['name_links'] = [
            link.find("a")['href'] for link in
            transfer.select("div.block-player")
            ]
            # Extract old and new team titles
            for key, class_name in [("old", "divCell Team OldTeam"),
                                     ("new", "divCell Team NewTeam")]:
                team_div = transfer.find("div", class_=class_name)
                a_tag = team_div.find("a") if team_div else None
                transfer_dict[key] = a_tag['title'] if a_tag else None
                if a_tag:
                    transfer_dict[f"{key}_link"] = a_tag['href']
            transfers_list.append(transfer_dict)
    return pd.DataFrame(transfers_list)



"""
Module with functions parsing multiple liquipedia objects with only one call

Dependencies
------------
- parse_liquipedia
- liquipedia_objects

Classes
-------
PageTypeRegistry
    Class defining mapping between strings and classes

Raises
------
MethodNotFoundException
    Raised when calling an invalid method on a liquipedia_object

"""

from typing import Callable, Optional, Union, List, Dict, Any
from ggpyparser.parse_liquipedia import parse_liquipedia_wc
from ggpyparser.liquipedia_objects import player, team, tournament

class MethodNotFoundException(Exception):
    """Exception raised when calling an invalid method on a liquipedia_object"""

class PageTypeRegistry:
    """Class defining mapping between strings and classes"""
    _registry = {}

    @classmethod
    def register(cls, type_name: str) -> Callable:
        """
        Registers a class under a given string type name.

        Parameters
        ----------
        type_name : str
            The string identifier for the class (e.g., "player", "team").
        
        Returns
        -------
        Callable
            A decorator that registers the decorated class under the specified type name.
        """
        def inner(page_class):
            cls._registry[type_name.lower()] = page_class
            return page_class
        return inner

    @classmethod
    def get_class(cls, type_name : str) -> Optional[type]:
        """
        Retrieves the class associated with the type_name
        """
        return cls._registry.get(type_name.lower())

# Register the actual classes, not subclasses or functions
PageTypeRegistry.register("tournament")(tournament.Tournament)
PageTypeRegistry.register("team")(team.Team)
PageTypeRegistry.register("player")(player.Player)

def create_multiple_pages(game : str,page_names : List[str], page_ts : Union[List[str], str],
                        user : str) -> Dict[str, Any]:
    """
        Helper function to get many pages in one api call - only wikicode parsing is supported
        Parameters
        ----------
        game: str
            The game being played
        page_names: List[str]
            The page names, found by liquipedia.com/game/(name)
        page_ts: Union[List[str], str]
            Types of pages being parsed
        user: str
            The user, as requested by liquipedia ToS
        
        
        Returns
        -------
            Dict[str, Any]
                A dictionary mapping page names to the corresponding liquipedia_page object
    """
    page_types = page_ts if isinstance(page_ts, list) else [page_ts.lower()] * len(page_names)
    response = parse_liquipedia_wc.make_request(
            user, game, "|".join(page_names), action = "wikicode"
        )
    objects = {}
    for name, ptype in zip(page_names, page_types):
        page_class = PageTypeRegistry.get_class(ptype)
        if page_class is None:
            raise ValueError(f"Page class for type '{ptype}' is not registered.")
        raw_str = response[name.lower().strip()]
        obj = page_class.from_raw_str(raw_str, user, game, name, action="wikicode")
        objects[name] = obj
    return objects



"""Testing the Player class"""
import unittest
from pathlib import Path
import pandas as pd
from ggpyscraper.liquipedia_objects import player
from ggpyscraper import parse_multiple_liquipedia_pages

GAME = "counterstrike"
PLAYERS = ['Autimatic', 'Stewie2K']
INFOS_wc = dict(zip(PLAYERS, [22, 24]))
INFOS_html = dict(zip(PLAYERS, [10, 11]))

GEAR_html = dict(zip(PLAYERS, [1,3]))

GEAR_wc = dict(zip(PLAYERS, [2,4]))
GEAR_CROSSHAIR_html = dict(zip(PLAYERS, [9,9]))

GEAR_CROSSHAIR_wc = dict(zip(PLAYERS, [3,3]))

TEAM_HISTORY = dict(zip(PLAYERS, [13, 12]))

ACHIEVEMENTS = dict(zip(PLAYERS, [10, 10]))

BASE_PATH = Path("tests/assets/player")
class TestPlayers(unittest.TestCase):
    """Test the Player class"""

    @classmethod
    def setUpClass(cls):
        # Build wikicode-backed Tournament objects (one network call batching page names)
        cls.all_players_wikicode = parse_multiple_liquipedia_pages.create_multiple_pages(
            game=GAME, page_names=PLAYERS, page_ts="player",
            user = "ggpyparser testing(github.com/Lou-Zhou)"
        )

        # Build HTML-backed player objects from local fixtures
        cls.all_players_html = {}
        for t_name in PLAYERS:
            path = BASE_PATH / f"{t_name.replace('/', '_')}.txt"
            with path.open("r", encoding="utf-8") as f:
                raw_html = f.read()
            t_obj = player.Player.from_raw_str(
                name=t_name, game=GAME, action="html", response=raw_html,
                user = "ggpyparser testing(github.com/Lou-Zhou)"
            )
            cls.all_players_html[t_name] = t_obj
    def setup_test_players(self, players,
                               ground_truths, method_name, mode="wikicode", both = False,
                               key = None):
        """Sets up and runs tests for players."""
        player_dict = self.all_players_wikicode if mode == "wikicode" else self.all_players_html
        results_tournament = {name: t_obj for name, t_obj in
                            player_dict.items() if
                            name in players}

        for to_name, to_obj in results_tournament.items():
            with self.subTest(name=to_name, mode=mode):
                method = getattr(to_obj, method_name)
                results = method()
                if key is not None:
                    results = len(results.get(key))
                elif isinstance(results, list):
                    results = len(results)
                elif isinstance(results, dict):
                    results = len(results)
                elif isinstance(results, pd.DataFrame):
                    results = results.shape if both else results.shape[0]
                expected_size = ground_truths[to_name]
                self.assertEqual(
                    results,
                    expected_size,
                    msg=f"{to_name}, {mode}",
                )
    def test_info_wc(self):
        """Validate info counts using WC parse."""
        self.setup_test_players(PLAYERS, INFOS_wc,
                                    "get_info", mode="wikicode")
    def test_info_html(self):
        """Validate info counts using HTML parse."""
        self.setup_test_players(PLAYERS, INFOS_html,
                                    "get_info", mode="html")
    def test_gear_html(self):
        """Validate gear counts using HTML parse."""
        self.setup_test_players(PLAYERS, GEAR_html, "get_gear", mode="html")
    def test_gear_wc(self):
        """Validate gear counts using WC parse."""
        self.setup_test_players(PLAYERS, GEAR_wc, "get_gear", mode="wikicode")
    def test_gear_crosshair_html(self):
        """Validate gear crosshair counts using HTML parse."""
        self.setup_test_players(PLAYERS, GEAR_CROSSHAIR_html, "get_gear",
                                mode="html", key = "Crosshair Settings")
    def test_gear_crosshair_wikicode(self):
        """Validate gear crosshair counts using wikicode parse."""
        self.setup_test_players(PLAYERS, GEAR_CROSSHAIR_wc, "get_gear",
                                 mode="wikicode", key = "Crosshair")
    def test_team_history_wc(self):
        """Validate team history counts using WC parse."""
        self.setup_test_players(PLAYERS, TEAM_HISTORY, "get_info",
                                mode="wikicode", key = "team_history")
    def test_team_history_html(self):
        """Validate team history counts using HTML parse."""
        self.setup_test_players(PLAYERS, TEAM_HISTORY, "get_info",
                                mode="html", key = "team_history")
    def test_achievements_html(self):
        """Validate achievements counts using HTML parse."""
        self.setup_test_players(PLAYERS, ACHIEVEMENTS, "get_achievements", mode="html")
if __name__ == "__main__":
    unittest.main()



"""
Module for representing and extracting information about a tournament.

Defines the Tournament class, provides
methods to fetch and parse tournament data like the prize pool

Classes
-------
Tournament
    Represents an individual tournament and provides methods to extract structured information
    about the competition

Dependencies
------------
- BeautifulSoup
- pandas
- re
- mwparserfromhell: For parsing MediaWiki markup.
- parse_liquipedia, parse_liquipedia_html, Custom modules for parsing specific data blocks.
- LiquipediaPage: Base class for interacting with Liquipedia.

"""
from typing import List, Union
from collections import defaultdict
import warnings
import re
import mwparserfromhell as mw
import pandas as pd

from bs4 import BeautifulSoup
from bs4.element import Tag
from ggpyparser.parse_liquipedia import parse_liquipedia_html, parse_liquipedia_wc
from ggpyparser.liquipedia_objects import liquipedia_page
class Tournament(liquipedia_page.LiquipediaPage):
    """
    A class to represent a tournament


    Methods
    -------
    get_info(infobox_name = "Infobox team")
        Parses information from tournament's infobox
    get_results()
        Get tournament results
    get_participants()
        Get all participants
    get_talent()
        Get all talent
    get_prizes()
        Get the tournament's prize pool
    """
    def __init__(self, game : str, name : str,
                 user : str,  action : str = "wikicode"
                 ) -> None:
        """
        Creates a tournament object

        Parameters
        ----------
        game: str
            The game being played
        name: str
            The page name, found by liquipedia.com/game/(name)
        user: str
            The user, as requested by liquipedia ToS
        action: str
            Whether html(action = "parse") or wikicode(action = "wikicode") parsing should occur
        """
        super().__init__(game, name, user=user, action = action)

    def _get_matches_wc(self) -> pd.DataFrame:
        """A private function to parse matches with wikicode"""
        sections = []
        parses = []
        str_parsed = mw.parse(self.raw_str)
        headings = [heading.title.strip().lower()  for heading in str_parsed.filter_headings()]
        results_sections = ['results', "result"] if "results" in headings else ["group stage",
                                                                      "playoffs", "main event"]
        for section in str_parsed.get_sections(include_lead=True, include_headings=True):
            if len(section.filter_headings()) == 0:
                continue
            #exists probably a cleaner solution to find relevant headers
            heading = section.filter_headings()[0].title.strip().lower()
            #find all results sections
            if heading in results_sections:
                #results = section.get_sections(include_lead=False, include_headings=False)
                parse = mw.parse(section)
                parses.append(parse) #incase of multiple "results" sections
        games_df = []
        #for all the results sections
        for parse in parses:
            sections = parse.get_sections(include_lead=False, include_headings=True)
            # Handle single-section tournaments (no subsections)
            if not sections:
                stage = "playoffs" if "Bracket" in parse else "group_stage"
                new_games = parse_liquipedia_wc.parse_games(stage, parse)
                new_games['stage_group'] = stage
                games_df.append(new_games)
                continue
            seen_stages = []
            lowest_sections = parse_liquipedia_wc.get_lowest_subsections(parse)
            for lowest in lowest_sections:
                stage = lowest.filter_headings()[0].title.strip().lower()
                if str(lowest) in seen_stages or "{{Match" not in str(lowest):
                    continue
                seen_stages.append(str(lowest))
                match = re.search(r"\{\{[^|]+\|([^}]+)\}\}", stage, re.I)
                if match:
                    stage = match.group(1).strip().lower()
                for low in lowest.split("{{Stage|"):
                    if "{{Match" not in low:
                        continue
                    new_games = parse_liquipedia_wc.parse_games(stage, low)
                    new_games['stage_group'] = stage
                    games_df.append(new_games)
        matches = pd.concat(games_df, ignore_index=True)
        if len(parses) > 0 and len(matches) == 0:
            raise parse_liquipedia_wc.SectionNotFoundException(
            "Found results section but could not parse game data, it is likely" \
            " that the game data is not in the wikicode for this tournament")
        if len(games_df) == 0:
            raise parse_liquipedia_wc.SectionNotFoundException(
            "Could not find the results section on the page, " \
            "ensure that the page has a results" \
            " section. If this tournament has stages, " \
            "it is likely that the results section is in the stage pages")
        return matches

    def _get_matches_html(self) -> pd.DataFrame:
        """A private function to parse matches with html"""
        match_idx = 0
        souped = BeautifulSoup(self.get_raw_str(), "html.parser")
        matchlists = souped.select(
        "div.general-collapsible.brkts-matchlist"
        )
        all_matches = []
        for matchlist in matchlists:
            header = matchlist.find_previous(["h1", "h2", "h3", "h4", "h5", "h6"])
            header = header.get_text().split("[")[0] if header else None
            title = matchlist.select(
                "div.brkts-matchlist-title")[0].get_text().replace(" Show Hide", "")
            matches = matchlist.find_all("div", class_ = "brkts-popup brkts-match-info-popup")
            for match in matches:
                parsed = parse_liquipedia_html.parse_match_html(match)
                parsed['stage'] = header
                parsed['substage'] = title
                parsed['match_idx'] = match_idx
                match_idx += 1
                all_matches.append(parsed)
        #parse brackets
        brackets = souped.select('div[class="brkts-bracket"]')
        for b_round in brackets:
            header = b_round.find_previous(["h1", "h2", "h3", "h4", "h5", "h6"])
            header = header.get_text().split("[")[0] if header else None
            subround_names = b_round.select('div[class="brkts-header brkts-header-div"]')
            subbrackets = b_round.find_all("div", class_="brkts-round-body", recursive = False)
            for subbracket in subbrackets:
                closest_header = subbracket.find_previous("div", class_="brkts-round-header")
                subround_names = [subround_name.contents[0]
                                  for subround_name in closest_header]
                subround_names = [name for name in subround_names
                                  if str(name).lower().strip() != "qualified"]
                output = parse_liquipedia_html.parse_bracket_recursive_html(
                    subbracket, subround_names)
                for sub_round, body in output.items():
                    for match in body:
                        parsed = parse_liquipedia_html.parse_match_html(match)
                        parsed['stage'] = header
                        parsed['substage'] = sub_round
                        parsed['match_idx'] = match_idx
                        match_idx += 1
                        all_matches.append(parsed)
        #get single matches(mostly showmatches or
        #grand-finals where the finalists are weirdly determined)
        single_matches = souped.find_all(
            "div", class_ = "brkts-popup brkts-popup brkts-match-info-flat")
        for match in single_matches:
            header = match.find_previous("span", class_="mw-headline").get_text()
            parsed = parse_liquipedia_html.parse_match_html(match)
            parsed['stage'] = header
            parsed['substage'] = header
            all_matches.append(parsed)
        if len(all_matches) == 0:
            raise parse_liquipedia_wc.SectionNotFoundException("No matches found")
        all_matches = pd.concat(all_matches)
        if all_matches.shape[0] == 0:
            warnings.warn("Found a Matches Section but no actual match data was found. " \
            "If this tournament has not started yet this is expected.")
            #TODO: for tournaments yet to be played, still return structure of tournament
        return all_matches.reset_index(drop = True)

    def get_results(self) -> pd.DataFrame:
        """Gets the results of a tournament"""
        if self.action == "wikicode":
            return self._get_matches_wc()
        return self._get_matches_html()

    def get_participants(self) -> pd.DataFrame:
        """Gets the participants of a tournament"""
        if self.action == "wikicode":
            return self._get_participants_wc()
        return self._get_participants_html()

    def _get_participants_html(self) -> pd.DataFrame:
        """Private function to get participants with html parsing"""
        raw_str = BeautifulSoup(self.get_raw_str(), "html.parser")
        participants = []
        for team in raw_str.find_all("div", class_ = "teamcard toggle-area toggle-area-1"):
            team_dict = {}
            header = team.find_previous(["h1", "h2", "h3", "h4", "h5", "h6"])
            header = header.get_text().split("[")[0] if header else None
            team_dict['name'] = team.find("center").get_text() if team.find("center") else None
            nums = team.find("table",
                            class_ =  "wikitable wikitable-bordered list active").find_all("tr")
            team_dict['qualification'] = (qual := team.find("td",
                                        class_="teamcard-qualifier")) and qual.get_text()
            team_dict['stage_enter'] = header

            for num in nums:
                text = str(num.get_text()).split()
                team_dict[f"p{text[0]}"] = text[1]
            participants.append(team_dict)
        if len(participants) == 0:
            raise parse_liquipedia_wc.SectionNotFoundException(
                "Could not find participants section")
        return pd.DataFrame(participants).reset_index(drop = True)

    def _get_participants_wc(self) -> pd.DataFrame:
        """Private function to get participants with wikicode parsing"""
        str_parsed = mw.parse(self.get_raw_str())
        all_players = []
        for section in str_parsed.get_sections(matches=r"(?i)^participants$"):
            lowest_section = parse_liquipedia_wc.get_lowest_subsections(section)
            for lowest in lowest_section:
                participant_stage = lowest.filter_headings()[0].title.strip().lower()
                match_tpl = [t for t in lowest.filter_templates(recursive=True)
                        if t.name.matches("ParticipantSection") or t.name.matches('TeamCard')]
                for participant in match_tpl:
                    p_dict = {"stage": participant_stage}
                    for param in participant.params:
                        text = str(param.value)
                        matches = re.findall(r"\[\[(.*?)\]\]", text)
                        if matches:
                            text = matches[0]
                        p_dict[str(param.name)] = re.sub(r"[{}\[\]]", "", text).split("|")[-1]
                    all_players.append(p_dict)
        return pd.DataFrame(all_players)

    def get_talent(self) -> pd.DataFrame:
        """Gets the talent of a tournament"""
        if self.action == "wikicode":
            return self._get_talent_wc()
        return self._get_talent_html()

    def _get_talent_html(self) -> pd.DataFrame:
        """Private function to get participants with html parsing"""
            #no explitict talent section in html, have to scan for ids
        souped = BeautifulSoup(self.get_raw_str(), "html.parser")
        all_talent = []
        talent_section = talent_span.parent if (talent_span :=
                                souped.find("span", {"id": "Broadcast_Talent"})) else None
        #could change depending on title, look in the future
        talent_container = talent_section.find_all("div",
                                    class_="tabs-content") if talent_section else None
        template_boxes = []
        if not talent_container:
            talent_container = talent_section.find_next_sibling("div") if talent_section else None
        if isinstance(talent_container, Tag):
            template_boxes = [
                    box for box in talent_container.find_all("div", class_="template-box")
                    if not box.find("div", class_="template-box")  # deepest level check
            ] + [box for box in talent_container.find_all("ul")]
        if not template_boxes and talent_section:
            template_boxes = list(filter(None, [talent_section.find_next("ul")]))
        for role_box in template_boxes:
            role_title = role_box.find("b")
            if role_title:
                if not isinstance(role_box, Tag) or not isinstance(role_title, Tag):
                    raise parse_liquipedia_wc.SectionNotFoundException(
                        "Unable to parse role box, not a bs4 Tag")
                role_name = role_title.get_text(strip=True).rstrip(":")
                for li in role_box.find_all("li"):
                    text = li.get_text().split("\n")
                    if len(text) > 1:
                        continue
                    flag = li.find("a")
                    country = flag['title'] if flag else None
                    talent_roles = {}
                    match = re.search(r'\xa0([^\xa0(]+)\xa0\(([^)]+)\)', text[0])
                    if match:
                        talent_roles['displayname'] = match.group(1).strip()
                        talent_roles['fullname'] = match.group(2).strip()
                        talent_roles['country'] = country
                    talent_roles["role"] = role_name
                    all_talent.append(talent_roles)
        if len(all_talent) == 0:
            raise parse_liquipedia_wc.SectionNotFoundException("Could not find talent section")
        all_talent = pd.DataFrame(all_talent)
        if "fullname" in all_talent.columns:
            #band-aid fix since the talent can be sometimes be parsed multiple times
            #this usually happens with tournaments with different stages and
            # if the talent section a mix of tabs and non-tabs
            return all_talent.drop_duplicates(subset="fullname", keep='last',
                                               inplace=False).reset_index(drop = True)
        all_talent = all_talent.drop_duplicates(subset = "role", inplace = False)
        if len(all_talent.columns) == 1:
            #if not announced, manually fill with TBA
            #this might cause some side-effects tho, so look out in future
            all_talent['fullname'] = "TBA"
        return all_talent.reset_index(drop = True)




    def _get_talent_wc(self) -> pd.DataFrame:
        """Private function to get participants with wikicode parsing"""
        #TODO: add language of talent
        talent_stage = 1
        broadcast_df = []
        str_parsed = mw.parse(self.raw_str)
        for section in str_parsed.get_sections(
            matches=r"(?i)^(broadcast talent|talent)$",
            include_lead=False,
            include_headings=True
        ):
            parse = mw.parse(section)
            #TODO: change to template matching instead of regex
            pattern = r"\{\{[Bb]roadcasterCard.*?\}\}"
            roles = re.findall(pattern, str(parse), re.DOTALL)
            for role in roles:
                role_df = parse_liquipedia_wc.parse_talent(role)
                role_df['talent_stage'] = talent_stage
                broadcast_df.append(role_df)
            talent_stage += 1
        if len(broadcast_df) == 0:
            raise parse_liquipedia_wc.SectionNotFoundException(
            "Could not find talent section on the page," \
            " ensure that the page has a talent" \
            "section. If this is a stage of a larger tournament, " \
            "it is likely that the talent section is in the tournament overview")
        return pd.concat(broadcast_df).reset_index(drop = True)
        #because of ul and div mixing for talent,
        #we need to do both which leads to double counting occassionally
        #might be a fix but seems difficult without sacrificing not counting some

    def get_prizes(self) -> Union[List[pd.DataFrame], pd.DataFrame]:
        """Gets the prize pool for a tournament"""
        if self.action == "wikicode":
            return self._get_prizes_wc()
        return self._get_prizes_html()

    def _get_prizes_html(self) -> Union[List[pd.DataFrame], pd.DataFrame]:
        """Private function to get prize pool with html parsing"""
        souped = BeautifulSoup(self.get_raw_str(), "html.parser")
        prize_pools = souped.find_all("div",
                    class_ = "csstable-widget collapsed general-collapsible prizepooltable")
        df_list = []
        for prize_pool in prize_pools:
            header = prize_pool.find("div", class_ = "csstable-widget-row prizepooltable-header")
            headers = [h.get_text(strip=True).lower() for h in
                       header.select(".csstable-widget-cell")]
            rows = []

            col_map = {idx: col_name for idx, col_name in enumerate(headers)}
            prize_rows = prize_pool.select(
            "div.csstable-widget-row:not(.prizepooltable-header):not(.ppt-toggle-expand)")
            for row in prize_rows:
                row_dict = defaultdict(list)
                cells = row.select(".csstable-widget-cell")
                teams = row.find_all("div", class_ = "block-team")
                teams =  [team.get_text() for team in teams]

                current_idx = 0
                span_idx = 0
                non_spans = []
                #check for spans
                spans = [cell for cell in cells if cell.get("style") and
                          "span" in cell.get("style") and
                          int(re.search(r"span (\d+)", cell["style"]).group(1)) > 1]

                for cell in cells:
                    if len(spans) > 0 and (not cell.get("style") or "span"
                                           not in cell.get("style")) and current_idx < len(col_map):
                        non_spans.append(current_idx)
                    if current_idx >= len(col_map):
                        new_idx = non_spans[span_idx %
                                            len(non_spans)] if len(non_spans) > 0 else new_idx
                        span_idx += 1
                        colname = col_map[new_idx]
                        row_dict[colname].append(cell.get_text(strip=True))
                    #print(cell)
                    else:
                        colname = col_map[current_idx]
                        row_dict[colname].append(cell.get_text(strip=True))
                        current_idx += 1
                    #colname = col_map[idx]
                    #print(cell.get_text())
                row_dict = {k: v[0] if len(v) == 1 else v for k, v in row_dict.items()}
                rows.append(row_dict)
            df_list.append(pd.DataFrame(rows).reset_index(drop = True))
        if len(df_list) == 0:
            raise parse_liquipedia_wc.SectionNotFoundException("Could not find prizes section")
        return df_list[0] if len(df_list) == 1 else df_list

    def _get_prizes_wc(self) -> pd.DataFrame:
        """Private function to get prize pool with wikicode parsing"""
        prizes = []
        str_parsed = mw.parse(self.raw_str)
        for section in str_parsed.get_sections(matches=r"(?i)^prize pool"):
            if "prize pool start" in str(section):
                prize_df = parse_liquipedia_wc.parse_prizes(section, match_1 = "prize pool slot",
                                                            match_2 = r"(?i)prize pool start")
            else:
                prize_df = parse_liquipedia_wc.parse_prizes(section)
            prizes.append(prize_df)

        return pd.concat(prizes).reset_index(drop = True)



"""
Module for representing and extracting information about a team.

Defines the Team class, provides
methods to fetch and parse team data like achievements

Classes
-------
Team
    Represents an individual team and provides methods to extract structured information
    about competitive history and news

Dependencies
------------
- BeautifulSoup
- pandas
- mwparserfromhell: For parsing MediaWiki markup.
- parse_liquipedia, parse_liquipedia_html, Custom modules for parsing specific data blocks.
- LiquipediaPage: Base class for interacting with Liquipedia.
"""
from typing import Dict, List, Union
import re
import pandas as pd
from bs4 import BeautifulSoup
import mwparserfromhell as mw
from ggpyparser.parse_liquipedia import parse_liquipedia_html, parse_liquipedia_wc
from ggpyparser.liquipedia_objects import liquipedia_page
ParsedValue = Union[str, pd.DataFrame, List[Dict[str, str]]]
class Team(liquipedia_page.LiquipediaPage):
    """
    A class to represent a liquipedia page.


    Methods
    -------
    get_info(infobox_name = "Infobox team")
        Parses information from the players infobox
    get_news()
        Get all news about the team
    get_players()
        Get all current and inactive players
    get_organization()
        Get all current and inactive organization members
    get_results()
        Get both the achievements and recent matches for a team
    """
    def __init__(self, game, name, user,
                  action = "wikicode"):
        """
        Creates a Team object

        Parameters
        ----------
        game: str
            The game being played
        name: str
            The page name, found by liquipedia.com/game/(name)
        user: str
            The user, as requested by liquipedia ToS
        action: str
            Whether html(action = "parse") or wikicode(action = "wikicode") parsing should occur
        """
        super().__init__(game, name, user=user, action = action)
    def get_info(self, infobox_name: str = "Infobox team"
                 ) -> Dict[str, Union[str, pd.DataFrame, List[Dict[str, str]]]]:
        """
        Gets the information from a team's infobox
        """
        info_dict =  super().get_info(infobox_name)
        for entry, text in info_dict.items():
            pattern = (
                r"\{\{Flag\|(\w+)\}\}\s*"
                r"(?:"
                    r"\[\[[^\|\]]+\|([^\]]+)\]\]"  # [[link|display]]
                    r"|"                           # or
                    r"\[\[([^\]]+)\]\]"            # [[link]]
                    r"|"                           # or
                    r"([A-Za-z0-9\-\s]+)"          # plain text fallback
                r")"
            )

            matches = re.findall(pattern, str(text))
            if len(matches) > 0:
                matches = [tuple(x for x in t if x) for t in matches]
                info_dict[entry] = [{"country": t1, "name": t2} for t1, t2 in matches]
        return info_dict

    def get_news(self) -> pd.DataFrame:
        """ Gets the data about a team's news """
        if self.action == "wikicode":
            return self._get_news_wc()
        return self._get_news_html()

    def _get_news_html(self) -> pd.DataFrame:
        """ Private method to get the data about a team's news using an html parse """
        souped = BeautifulSoup(self.get_raw_str(), "html.parser")
        timeline_data = parse_liquipedia_html.get_all_under_header(souped, "Timeline")
        tabbed_data = []
        for data in timeline_data:
            if data.get("class") and  any("tabs" in c for c in data.get("class")):
                timeline_table = parse_liquipedia_html.build_tab_map(data)
                tabbed_data.append(pd.concat
                                   ([pd.DataFrame(parse_liquipedia_html.
                                     parse_single_tab_history(text, year))
                                     for year, text in timeline_table.items()]))
            else:
                tabbed_data.append(pd.DataFrame(parse_liquipedia_html.
                                                parse_single_tab_history(data)))
        return pd.concat(tabbed_data).reset_index(drop = True)

    def _get_news_wc(self) -> pd.DataFrame:
        """ Private method to get the data about a team's news using wikicode parse """
        news_data = []
        parsed = mw.parse(self.raw_str)
        for section in parsed.get_sections(include_lead=False, include_headings=True):
            header = section.filter_headings()[0].title.strip().lower()
            if header == "timeline":
                year_text_map = parse_liquipedia_wc.get_name_content_map(str(section))
                for year, text in year_text_map.items():
                    entries = text.split("\n")
                    for entry in entries:
                        data = parse_liquipedia_wc.parse_news_str(entry)
                        if data != -1 and len(data) > 0:
                            data['year'] = year
                            news_data.append(data)
        return pd.DataFrame(news_data).reset_index(drop = True)

    def get_players(self) -> Union[pd.DataFrame, Dict[str, pd.DataFrame]]:
        """Gets information about a team's players"""
        if self.action == "wikicode":
            return self._get_people_wc("player roster")
        return self._get_people_html(header = "Player_Roster")

    def get_organization(self)  -> Union[pd.DataFrame, Dict[str, pd.DataFrame]]:
        """Gets information about a team's organization"""
        if self.action == "wikicode":
            return self._get_people_wc("organization")
        return self._get_people_html(header = "Organization")

    def _get_people_html(self, header) -> Union[pd.DataFrame, Dict[str, pd.DataFrame]]:
        """Private method parsing a section describing people with html
        Parameters
        ----------
            header: str
                The name of the header of interest

        Returns
        -------
            Dict[str, pd.DataFrame]]
                A dataframe or dictionary of dataframes describing 
                information about the section's people
        """
    #game.find_all("div", class_ = "table-responsive")
        souped = BeautifulSoup(self.get_raw_str(), "html.parser")
        all_players = []
        for section in parse_liquipedia_html.get_all_under_header(souped, header = header):
            tab_map = parse_liquipedia_html.build_tab_map(section)
            if len(tab_map) == 0:
                all_players = all_players + parse_liquipedia_html.parse_players_raw(
                    section, self.game)
            else:
                for game, text in tab_map.items():
                    all_players = all_players + parse_liquipedia_html.parse_players_raw(
                        text, game)
        return pd.concat(all_players).reset_index(drop = True)

    def _get_people_wc(self, header) -> Union[pd.DataFrame, Dict[str, pd.DataFrame]]:
        """Private method parsing a section describing people with wikicode
        Parameters
        ----------
            header: str
                The name of the header of interest

        Returns
        -------
            Dict[str, pd.DataFrame]]
                A dataframe or dictionary of dataframes describing 
                information about the section's people
        """
        all_people = []
        parsed = mw.parse(self.raw_str)
        for section in parsed.get_sections(include_lead=False, include_headings=True):
            sec_title = section.filter_headings()[0].title.strip().lower()
            if sec_title == header:
                #get players(non standins)
                p_tpl = [p for p in section.filter_templates(recursive=True)
                 if p.name.matches("Person") or p.name.matches("stand-in")]
                for player in p_tpl:
                    player_dict = {}
                    for param in player.params:
                        entry = str(param.value)
                        entry = re.sub(r"<ref.*?(.*?)/>", "", entry)
                        entry = re.sub(r"<ref.*?>(.*?)</ref>", "", entry)
                        matches = re.findall(r"\[\[(.*?)\]\]", entry)
                        if matches:
                            entry = matches
                        player_dict[str(param.name)] = entry

                    all_people.append(player_dict)
        return pd.DataFrame(all_people).reset_index(drop = True)
    def get_results(self):
        """Parses results section for a team"""
        if self.action == "wikicode":
            raise parse_liquipedia_wc.SectionNotFoundException(
                "Cannot parse results section using action = query, try action = parse")

        souped = BeautifulSoup(self.get_raw_str(), "html.parser")
        timeline_data = parse_liquipedia_html.get_all_under_header(souped, "Results")
        for data in timeline_data:
            if data.get("class") and  any("tabs" in c for c in data.get("class")):
                timeline_table = parse_liquipedia_html.build_tab_map(data)
                return {k: parse_liquipedia_html.
                        parse_wikitable_achievements(v).
                        reset_index(drop = True) for k,v in timeline_table.items()}
        raise parse_liquipedia_wc.SectionNotFoundException("Could not find results table")
    


"""
Module for representing and extracting information about a general liquipedia page.

Defines the liquipedia page class, provides
methods to fetch and parse general page data such as the infobox

Classes
-------
LiquipediaPage
    Represents an individual page and provides methods to extract structured information
    about the page's infobox and other helper functions

Dependencies
------------
- BeautifulSoup
- pandas
- mwparserfromhell: For parsing MediaWiki markup.
- parse_liquipedia, Custom module for parsing specific data blocks.

Raises
------
UnknownParsingMethodException
    Raised when the parsing method is not "action" or "wikicode"
"""
from typing import Dict, Union,Optional, Type, TypeVar, List
from urllib.parse import unquote
import re
import warnings
import mwparserfromhell as mw
from bs4 import BeautifulSoup
from bs4.element import Tag


import pandas as pd
from ggpyparser.parse_liquipedia import parse_liquipedia_wc
T = TypeVar('T', bound='LiquipediaPage')
class UnknownParsingMethodException(Exception):
    """Exception raised when the parsing method is not html or wikicode"""
class LiquipediaPage:
    """
    A class to represent a liquipedia page.

    ...

    Attributes
    ----------
    game: str
        The game being played
    name: str
        The page name, found by liquipedia.com/game/(name)
    user: str
        The user, as requested by liquipedia ToS
    action: str
        Whether html(action = "parse") or wikicode(action = "wikicode") parsing should occur
    raw_str: str
        The raw string describing the text of the player's page

    Methods
    -------
    get_info(infobox_name = "Infobox player")
        Parses information from the players infobox
    get_raw_str()
        Gets the raw string describing the page's text
    """
    def __init__(self, game : str, name : str,
                user : str,
                action : str= "wikicode") -> None:
        """
        Creates a LiquipediaPage object

        Parameters
        ----------
        game: str
            The game being played
        name: str
            The page name, found by liquipedia.com/game/(name)
        user: str
            The user, as requested by liquipedia ToS
        action: str
            What parsing method should be used:
                wikicode - for wikicode parsing
                html_1 - for html parsing using the "parse" action(30s rate limit, reliable)
                html_2 - for html parsing using the "wikicode" action(2s rate limit, less reliable)
        """
        if action not in ["wikicode", "html"]:
            raise UnknownParsingMethodException("Unknown Parsing Method")
        self.user = user
        self.game = game
        self.name = unquote(name)
        self.action = action
        self.raw_str = self._make_request()


    def get_raw_str(self):
        """Returns the page's raw text"""
        return self.raw_str


    def _make_request(self) -> str:
        """Makes the API call depending on whether to use html or wikicode parsing"""
        raw_str = list(parse_liquipedia_wc.make_request(self.user,
                                                    self.game,self.name, self.action).values())[0]
        if self.action == "wikicode":
            match = re.search(r"#REDIRECT\s*\[\[(.*?)\]\]", raw_str, flags=re.IGNORECASE)
            if match:
                #check if redirect is needed
                new_name = match.group(1)
                raw_str = list(parse_liquipedia_wc.make_request(self.user,
                                                    self.game, new_name, self.action).values())[0]

            return str(raw_str)
        else:
            souped = BeautifulSoup(raw_str, "html.parser")
            redirect = souped.find("div", class_ = "redirectMsg")
            redirect = redirect.find("a") if redirect else None
            if isinstance(redirect, Tag):
                new_name = redirect.get("href") if redirect else None
                if isinstance(new_name, str):
                    new_name = new_name.rsplit("/", 1)[1]
                    raw_str = list(parse_liquipedia_wc.make_request(self.user, self.game,
                                                                new_name, self.action).values())[0]
            return str(raw_str)

    def get_info(self, infobox_name: str = "Infobox league"
                 ) -> Dict[str, Union[str, pd.DataFrame, List[Dict[str, str]]]]:
        """Gets the information from the page's infobox """
        if self.action == "wikicode":
            return self._get_info_wc(infobox_name)
        return self._get_info_html()

    def _get_info_html(self) -> Dict[str, Union[str, pd.DataFrame, List[Dict[str, str]]]]:
        """Private method to get page's infobox using a html parse"""
        souped = BeautifulSoup(self.get_raw_str(), "html.parser")
        infoboxes = souped.select('div[class="fo-nttax-infobox"]')
        if len(infoboxes) == 0:
            raise parse_liquipedia_wc.SectionNotFoundException("Infobox Section not Found")
        if len(infoboxes) > 1:
            warnings.warn("Multiple infoboxes detected, taking first one found", UserWarning)
        infobox = infoboxes[0]

        rows = infobox.find_all("div", class_="infobox-cell-2 infobox-description")
        info = {}

        for row in rows:
            key = row.get_text(strip=True).rstrip(":")
            value_div = row.find_next_sibling("div")
            value = [a.get_text(strip=True) for a in value_div.find_all("a") if
                     a.get_text(strip=True)]
            if len(value) <= 1:
                value = value_div.get_text("", strip = True)
            info[key] = value
        return info



    def _get_info_wc(self, infobox_name:str
                     ) -> Dict[str, Union[str, pd.DataFrame, List[Dict[str, str]]]]:
        """Private method to get page's infobox using a wikicode parse"""
        infobox_dict = {}
        str_parsed = mw.parse(self.raw_str)
        for template in str_parsed.filter_templates():
            if template.name.matches(infobox_name):
                for param in template.params:
                    key = str(param.name).strip()
                    value = str(param.value).strip().replace("<br>", ", ")
                    infobox_dict[key] = value
                self.name = infobox_dict['name'] if (self.name is None
                                                     and "name" in infobox_dict) else self.name
                break
        if len(infobox_dict) == 0:
            raise parse_liquipedia_wc.SectionNotFoundException(
                "Infobox Section not Found, If this is a substage of a larger tournament, " \
                "the infobox is probably in the tournament page")
        return infobox_dict
    @classmethod
    def from_raw_str(
        cls: Type[T],
        response: str,
        user: str,
        game: Optional[str] = None,
        name: Optional[str] = None,
        action: str = "wikicode"
    ) -> T:
        """Alternate constructor to build a liquipedia page object from the raw text"""
        obj = cls.__new__(cls)
        obj.user = user
        obj.raw_str = response
        obj.game = game
        obj.name = unquote(name)
        if action not in ["wikicode", "html"]:
            raise UnknownParsingMethodException("Unknown Parsing Method")
        obj.action = action
        return obj



"""Testing the team class"""
import unittest
from pathlib import Path
import pandas as pd
from ggpyscraper.liquipedia_objects import team
from ggpyscraper import parse_multiple_liquipedia_pages

GAME = "counterstrike"
TEAMS = ['Evil_Geniuses.ca', 'G2_Esports']
test = team.Team(game = "counterstrike", name = "G2_Esports",
                 action = "html", user = "ggpyparser-testing(github.com/Lou-Zhou)")
#get_info, news, roster, organization, results(achievements), results(recent matches)
#WC: 26, 70, 69, 24, none, none
#HTML: 10, 70, 69, 24, 10, 10

#get_info, news, roster, organization, results(achievements), results(recent matches)
#WC: 12, 2, 9, 1, none, none
#HTML: 6, 2, 9, 1, 4, 10
INFOS_wc = dict(zip(TEAMS, [12, 26]))
INFOS_html = dict(zip(TEAMS, [6, 10]))

NEWS = dict(zip(TEAMS, [2, 71]))

ROSTER = dict(zip(TEAMS, [9, 69]))
ORGANIZATION = dict(zip(TEAMS, [1, 24]))
ACHIEVEMENTS = dict(zip(TEAMS, [4, 10]))
RECENT_MATCHES = dict(zip(TEAMS, [10, 10]))


BASE_PATH = Path("tests/assets/team")
class Testteams(unittest.TestCase):
    """Test the team class"""

    @classmethod
    def setUpClass(cls):
        # Build wikicode-backed Tournament objects (one network call batching page names)
        cls.all_teams_wikicode = parse_multiple_liquipedia_pages.create_multiple_pages(
            game=GAME, page_names=TEAMS, page_ts="team",
            user = "ggpyparser-testing(github.com/Lou-Zhou)"
        )

        # Build HTML-backed team objects from local fixtures
        cls.all_teams_html = {}
        for t_name in TEAMS:
            path = BASE_PATH / f"{t_name.replace('/', '_')}.txt"
            with path.open("r", encoding="utf-8") as f:
                raw_html = f.read()
            t_obj = team.Team.from_raw_str(
                name=t_name, game=GAME, action="html",
                response=raw_html, user = "ggpyparser-testing(github.com/Lou-Zhou)"
            )
            cls.all_teams_html[t_name] = t_obj
    def setup_test_teams(self, teams,
                               ground_truths, method_name, mode="wikicode", both = False,
                               key = None):
        """Sets up and runs tests for TEAMS."""
        team_dict = self.all_teams_wikicode if mode == "wikicode" else self.all_teams_html
        results_tournament = {name: t_obj for name, t_obj in
                            team_dict.items() if
                            name in teams}

        for to_name, to_obj in results_tournament.items():
            with self.subTest(name=to_name, mode=mode):
                method = getattr(to_obj, method_name)
                results = method()
                if key is not None:
                    results = len(results.get(key))
                elif isinstance(results, list):
                    results = len(results)
                elif isinstance(results, dict):
                    results = len(results)
                elif isinstance(results, pd.DataFrame):
                    results = results.shape if both else results.shape[0]
                expected_size = ground_truths[to_name]
                self.assertEqual(
                    results,
                    expected_size,
                    msg=f"{to_name}, {mode}",
                )
    def test_get_info_html(self):
        """Tests the get_info method for HTML"""
        self.setup_test_teams(TEAMS, INFOS_html, "get_info", mode="html")
    def test_get_info_wikicode(self):
        """Tests the get_info method for wikicode"""
        self.setup_test_teams(TEAMS, INFOS_wc, "get_info", mode="wikicode")
    #OUTDATED:
    #def test_get_news_html(self):
    #    """Tests the get_news method for HTML"""
        self.setup_test_teams(TEAMS, NEWS, "get_news", mode="html")
    def test_get_news_wikicode(self):
        """Tests the get_news method for wikicode"""
        self.setup_test_teams(TEAMS, NEWS, "get_news", mode="wikicode")
    def test_get_roster_html(self):
        """Tests the get_players method for HTML"""
        self.setup_test_teams(TEAMS, ROSTER, "get_players", mode="html")
    def test_get_roster_wikicode(self):
        """Tests the get_players method for wikicode"""
        self.setup_test_teams(TEAMS, ROSTER, "get_players", mode="wikicode")
    def test_get_organization_html(self):
        """Tests the get_organization method for HTML"""
        self.setup_test_teams(TEAMS, ORGANIZATION, "get_organization", mode="html")
    def test_get_organization_wikicode(self):
        """Tests the get_organization method for wikicode"""
        self.setup_test_teams(TEAMS, ORGANIZATION, "get_organization", mode="wikicode")
    def test_get_achievements_html(self):
        """Tests the get_results method for HTML"""
        self.setup_test_teams(TEAMS, ACHIEVEMENTS, "get_results", mode="html",
                              both = True, key = "Achievements")
    def test_get_recent_matches_html(self):
        """Tests the get_results method for HTML"""
        self.setup_test_teams(TEAMS, RECENT_MATCHES, "get_results", mode="html",
                              both = True, key = "Recent Matches")

if __name__ == "__main__":
    unittest.main()
